// Suggested improvements:
// - Interpolation should be nicer.  Optionally use texture mapping.
// - Add panning with arrow keys
// - Pan with mouse
// - Zoom with mouse
//
// History
// 4. Initial window size is now at least 300x300 pixels.
//    Pressing 'q' quits.
//    [ and ] keys now do x-zooming.
//    The window size is chosen more intelligently at start-up, as is the
//     scale factor for viewing the image.
//    The window dynamically resizes in an attempt to fit the size of the
//    image when it is rescaled.  This is handy.
// 3. Zooming has been added.
// 2. PGM images are now displayed correctly.  The previous version showed
// them all in blue.  Some files could not be opened before, but after moving
// glutInit() below the part where files are read in the bug went away.
// 1. 'p' and 'n' keys move to previous and next images, respectively.  Images
//    can be loaded from stdin or from arguments.  The window title is updated
//    whenever the user decides to view a different image.

#include <stdlib.h>
#include <pnm.h>
#include <assert.h>
#include <sys/types.h>
#include <unistd.h>
#include <math.h>
#include <string.h>
//#include <getopt_long.h>

#ifdef DARWIN
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif

#include <stdio.h>

#define MAX_IMAGES (1<<10)

#ifndef max
#define max(a,b) ((a)>(b)?(a):(b))
#endif

#ifndef min
#define min(a,b) ((a)<(b)?(a):(b))
#endif

typedef struct
{
    int w,h,format;
    xelval maxval;
    GLubyte *gl_pixels;
    char *filename;
}
image_t;

int using_texture=1;
image_t images[MAX_IMAGES];
int current_index=0;
int num_images=0;
int win_id=-1;
int winw=512;
int winh=512;
int minwinw=1024/2;
int minwinh=768/2;
int maxwinw=1024;
int maxwinh=768;
char *progname=NULL;
GLfloat zoom=1.0f,minzoom=1e-6f;
GLfloat xzoom=1.0f;

static void gl_init(void);
static void print_usage(void);
static void on_display(void);
static void on_keyboard(unsigned char key,int x, int y);
static void on_mouse(int button,int state,int x,int y);
static void on_reshape(int w,int h);
static void load_image(FILE* file,image_t* image);
static void update_window_title(void);
static void run_current_image_through_shell_pipe(void);
static void print_keys(void);
static void update_window_size(void);

int main(int argc,char *argv[])
{
    int i;

    pnm_init(&argc,argv);
    progname=argv[0];

    if((argv[1]==NULL)||(strcmp(argv[1],"-h")==0)||(strcmp(argv[1],"--help")==0)) {
        print_usage();
        exit(1);
    }

    for(i=1;i<argc;i++) {
        char* filename;
        FILE* file;

        filename=argv[i];
        if(strcmp(filename,"-")==0) {
            fprintf(stderr,"%s: Waiting for a PNM image on stdin...\n",progname);
            while(!feof(stdin)) {
                int got_eof;
                if(num_images==MAX_IMAGES) { 
                    fprintf(stderr,"Warning: Max number of images (%i) exceeded\n",
                            MAX_IMAGES);
                    break;
                }
                image_t* image=&images[num_images++];
                image->filename=NULL;
                load_image(stdin,image);
                pnm_nextimage(stdin,&got_eof);
                if(got_eof) { break; }
            }
        }
        else { 
            image_t* image=&images[num_images++];
            file=fopen(filename,"r");
            if(file==NULL) {
                fprintf(stderr,"%s: Could not open \"%s\" for reading\n",
                        progname,filename);
                exit(2);
            }
            image->filename=filename;
            load_image(file,image);
            fclose(file);
        }
    }

    if(num_images==0) {
        fprintf(stderr,"%s: No images read.\n",progname);
        exit(3);
    }

    image_t* img=&images[0];
    winw=min(maxwinw,max(minwinw,img->w));
    winh=min(maxwinh,max(minwinh,img->h));

    /* Set the default zoom factor so that the whole image fits in the window. */
    {
        GLfloat xfac,yfac;
        xfac=winw/(double)img->w;
        yfac=winh/(double)img->h;
        zoom=min(xfac,yfac);
        winw=min(winw,zoom*img->w);
        winh=min(winh,zoom*img->h);
    }

    
    glutInit(&argc,argv);
    glutInitDisplayMode(GLUT_DOUBLE|GLUT_RGB|GLUT_ALPHA);
    glutInitWindowSize(winw,winh);
    win_id=glutCreateWindow("pnm-glview");
    glutDisplayFunc(on_display);
    glutKeyboardFunc(on_keyboard);
    glutMouseFunc(on_mouse);
    glutReshapeFunc(on_reshape);
    gl_init();
    update_window_title();
    glutMainLoop();

    return 0;
}

static void print_usage(void)
{
        fprintf(stderr,
"\n"
"pnm-glview\n"
"Copyright (c) 2005, Issac Trotts\n"
"\n"
"View PNM format images with a GLUT-based OpenGL viewer.\n"
"\n"
"Examples\n"
"========\n"
"View an image from stdin:\n"
"\n"
"    cat image.ppm | pnm-glview -\n"
"\n"
"View multiple images from stdin:\n"
"    \n"
"    cat image1.ppm image2.pgm image3.ppm | pnm-glview -\n"
"\n"
"View some images specified on the command line:\n"
"\n"
"    pnm-glview image1.ppm image2.pgm image3.ppm \n"
"\n"
"Command line options\n"
"====================\n"
" -h : Display this help message.\n"
"\n");
    print_keys();
}

static void print_keys(void)
{
    fprintf(stderr,
"Keys\n"
"====\n"
"<ESC> : Quit.\n"
"?     : Print this message.\n"
"!     : Execute a shell command on the current image.  The command is assumed to \n"
"        read the image on stdin and output a new image on stdout.\n"
"q     : Quit.\n"
"p     : Display the previous image.\n"
"n     : Display the next image.\n"
"+     : Zoom in.\n"
"=     : Zoom in.\n"
"-     : Zoom out.\n"
"[     : Zoom out, x only.\n"
"]     : Zoom in, x only.\n"
"\n");
}

static void gl_init(void)
{
    glPixelStorei(GL_PACK_ALIGNMENT,1);
    glPixelStorei(GL_UNPACK_ALIGNMENT,1);
}

static void on_display(void)
{
    image_t* im;
    im=&images[current_index];
    glClear(GL_COLOR_BUFFER_BIT);
    glPixelZoom(xzoom*zoom,zoom);
    glDrawPixels(im->w,im->h,GL_RGBA,GL_UNSIGNED_BYTE,im->gl_pixels);
    glutSwapBuffers();
}

static void on_keyboard(unsigned char key,int x, int y)
{
    switch(key) {
        case 27:  exit(0); break;
        case 'Q': exit(0); break;
        case 'q': exit(0); break;

        case 'p':
            current_index--;
            if(current_index<0) { current_index=0; }
            else { glutPostRedisplay(); }
            update_window_title();
            break;
        case 'n':
            current_index++;
            if(current_index>=num_images) { current_index=num_images-1; }
            else { glutPostRedisplay(); }
            update_window_title();
            break;
        case '[':
            xzoom/=1.1f;
            if(xzoom<minzoom) { xzoom=minzoom; }
            update_window_size();
            break;
        case ']':
            xzoom*=1.1f;
            update_window_size();
            break;
        case '+':
        case '=':
            zoom*=1.1f;
            update_window_size();
            break;
        case '-':
            zoom/=1.1f;
            if(zoom<minzoom) { zoom=minzoom; }
            update_window_size();
            break;
        case '?':
            print_keys();
            break;
        case '!':
            run_current_image_through_shell_pipe();
            glutPostRedisplay();
            break;
        default:
            ;
    }
}

static void on_mouse(int button,int state,int x,int y)
{
    
}

static void on_reshape(int w,int h)
{
    glViewport(0,0,w,h);
    winw=w;
    winh=h;
    glOrtho(1,w,1,h,-10,10);
}

static void load_image(FILE* file,image_t* image)
{
    int ri,ci;
    const int nchan=3;
    xel *xelrow;

    pnm_readpnminit(file,&image->w,&image->h,&image->maxval,
            &image->format);
    image->gl_pixels=
        (GLubyte*)malloc(image->w*image->h*4*sizeof(GLubyte));
    xelrow=pnm_allocrow(image->w);
    for(ri=0;ri<image->h;ri++) {
        pnm_readpnmrow(file,xelrow,image->w,image->maxval,image->format);
        for(ci=0;ci<image->w;ci++) {
            int index=((image->h-1 - ri)*image->w+ci)*4;
            pixel p=PNM_FORMAT_TYPE(image->format)==PGM_FORMAT?
                xeltopixel(xelrow[ci]):xelrow[ci];
            image->gl_pixels[index+0]=PPM_GETR(p);
            image->gl_pixels[index+1]=PPM_GETG(p);
            image->gl_pixels[index+2]=PPM_GETB(p);
            image->gl_pixels[index+3]=255;
        }
    }
}

static void update_window_title(void)
{
    char title[80];

    if(images[current_index].filename==NULL) {
        snprintf(title,sizeof(title),"pnm-glview <stdin> (%i)",current_index);
    }
    else {
        snprintf(title,sizeof(title),"pnm-glview %s",images[current_index].filename);
    }
    glutSetWindowTitle(title);
}

static void run_current_image_through_shell_pipe(void)
{
    int i,j;
    int forceplain=0; // raw
    int maxval=255;
    char line[80*4],command[80*(4+1)+3],tmpname[80];
    pixel* pixelrow;
    fprintf(stderr,"Please enter a shell command to invoke on the current image.\n");
    fprintf(stderr,"> ");
    if(fgets(line,sizeof(line),stdin)==NULL) {
        fprintf(stderr,"EOF encountered.\n");
        return;
    }
    if(strlen(line)!=0) { line[strlen(line)-1]='\0'; } // Strip off trailing newline
    snprintf(tmpname,sizeof(tmpname),"/var/tmp/pnmglview%i.pnm",getpid());
    snprintf(command,sizeof(command),"%s > %s",line,tmpname);
    fprintf(stderr,"Command : %s \n",command);
    FILE* pipe=popen(command,"w");
    if(pipe==NULL) {
        fprintf(stderr,"Unable to open pipe for command \"%s\"\n",command);
        return;
    }
    image_t* img=&images[current_index];
    ppm_writeppminit(pipe,img->w,img->h,maxval,forceplain);
    pixelrow=ppm_allocrow(img->w);
    for(i=0;i<img->h;i++) { 
        for(j=0;j<img->w;j++) {
            pixel p;
            int index;
            index=((img->h-1-i)*img->w+j)*3;
            p.r=img->gl_pixels[index  ];
            p.g=img->gl_pixels[index+1];
            p.b=img->gl_pixels[index+2];
            pixelrow[j]=p;
        }
        ppm_writeppmrow(pipe,pixelrow,img->w,maxval,forceplain);
    }
    pclose(pipe);
    ppm_freerow(pixelrow);

    // Now read the results in from the temp file.
    free(img->gl_pixels);
    FILE* tempfile=fopen(tmpname,"rb");
    load_image(tempfile,img);
    fclose(tempfile);
}

static void update_window_size(void)
{
    winw=min(maxwinw,xzoom*zoom*images[0].w); 
    winh=min(maxwinh,zoom*images[0].h); 
    glutReshapeWindow(winw,winh);
    glutPostRedisplay();
}



