char* usage = 
"Usage: pgm-amplitude-spectrum image.pgm\n"
"The results are stored in spectrum.pgm.\n"
;

#include <stdio.h>
#include <stdlib.h>
#include <complex.h>
#include <pgm.h>
#include <fftw3.h>
#include <math.h>

#define FOREACH(dothis) \
    for(yi=0;yi<ny;yi++) { for(xi=0;xi<nx;xi++) { dothis; } } 

double cabs2(complex z) { return cabs(z*z); }

void fftshift2(complex in[], complex out[], int nx, int ny)
{
    int xi,yi;
    FOREACH(
        int y2=(yi+ny/2)%ny; 
        int x2=(xi+nx/2)%nx;
        out[yi*nx+xi] = in[y2*nx+x2];
        );
}

int main(int argc, char *argv[])
{
    gray maxval,outmaxval=255;
    gray** pixels;
    int xi,yi;
    int nx,ny;
    complex *in,*out;
    char* filename;

    pgm_init(&argc,argv);     
    if(argc!=1+1) {
        fprintf(stderr,usage);
        exit(-1);
    }
    filename=argv[1];
    
    // Load the image
    {
        FILE* file = fopen(filename,"rb");
        if(!file) {
            fprintf(stderr,"Could not open %s for reading.\n");
            fprintf(stderr,usage);
            exit(-1);
        }
        pixels = pgm_readpgm(file,&nx,&ny,&maxval);
        fclose(file);
    }

    // Compute the FFT
    fftw_complex* ft = fftw_malloc(sizeof(fftw_complex)*nx*ny); // shifted 
    {
        fftw_complex* out = fftw_malloc(sizeof(fftw_complex)*nx*ny);         
        fftw_complex* in = fftw_malloc(sizeof(fftw_complex)*ny*nx);         

        // Put the filter pixels into a flat array for fftw
        FOREACH(in[yi*nx+xi] = pixels[yi][xi]);
        
        fftw_plan plan = fftw_plan_dft_2d(ny, nx, in, out, FFTW_FORWARD,
                FFTW_ESTIMATE);
        fftw_execute(plan);
        fftw_destroy_plan(plan);
        fftw_free(in);

        // Divide out factor of w*h.
        FOREACH(out[yi*nx+xi] /= ny*nx);

        fftshift2(out,ft,nx,ny);
    }

    // Compute min, max, mean, sdev of amplitude spectrum
    double pmin,pmax;
    double pmean,psdev;
    {
        double pmeansq,pvar;
        pmin=pmax=cabs(ft[0]);
        pmean=pmeansq=0.0;
        FOREACH(
            complex val=ft[yi*nx+xi];
            pmin=MIN(pmin,cabs(val)); pmax=MAX(pmax,cabs(val));
            pmean+=cabs(val);
            pmeansq+=cabs(val)*cabs(val);
        );
        pmean/=nx*ny;
        pmeansq/=nx*ny;
        pvar=pmeansq-pmean*pmean;
        psdev=sqrt(pvar);
    }

    // Output the amplitude spectrum, putting the min and max values in a
    // comment at the top.
    {
        char* outname = "spectrum.pgm";
        FILE* outfile = fopen(outname,"wb");
        if(!outfile) {
            fprintf(stderr,"Could not open %s for writing.\n",outname);
            exit(-1);
        }
        fprintf(outfile,"P5\n");
        fprintf(outfile,"# Amplitude spectrum of %s\n",filename);
        fprintf(outfile,"%i %i\n",nx,ny);
        fprintf(outfile,"%i\n",outmaxval);
        FOREACH(
            double val = 128.0+128.0*(cabs(ft[yi*nx+xi])-pmean)/psdev;
            fputc((int)MIN(255,MAX(0,val)),outfile)
            );
        fclose(outfile);
        fprintf(stderr,"Wrote %s\n",outname);
    }

    // Convert to polar coordinates and integrate over theta.
    {
        char* apmname="amp-polar-marginals.txt";
        char* powname="pow-polar-marginals.txt";
        FILE* apmfile=fopen(apmname,"w");
        FILE* powfile=fopen(powname,"w");
        if(!apmfile) {
            fprintf(stderr,"Could not open %s for writing.\n",apmname);
            exit(-1);
        }
        if(!powfile) {
            fprintf(stderr,"Could not open %s for writing.\n",powname);
            exit(-1);
        }
        int t,fudge=2;
        double maxr=MIN(nx,ny)/2-fudge;
        double r;
        for(r=1;r<maxr;r+=0.1) {
            int maxt=500.0*r/maxr;
            double as=0.0,ps=0.0;
            for(t=0;t<maxt;t++) {
                double theta=(2.0*M_PI*t)/maxt;
                double x,y;
                x=nx/2+r*cos(theta);
                y=ny/2+r*sin(theta);
                
                // Bilerp to get the value here.
                double val;
                {
                    int x0,x1,y0,y1;
                    double u0,u1,v0,v1;
                    x0=x; x1=ceil(x);
                    y0=y; y1=ceil(y);
                    u1=x-x0; u0=1.0-u1;
                    v1=y-y0; v0=1.0-v1;
                    val=u0*v0*cabs(ft[x0+y0*nx])+
                        u1*v0*cabs(ft[x1+y0*nx])+
                        u0*v1*cabs(ft[x0+y1*nx])+
                        u1*v1*cabs(ft[x1+y1*nx]);
                }
                as+=val;
                ps+=val*val;
            }
            as/=maxt;
            ps/=maxt;
            fprintf(apmfile,"%f\n",as);
            fprintf(powfile,"%f\n",ps);
        }
        fprintf(stderr,"Wrote %s and %s\n",apmname,powname);
    }

    return 0;
}

