char* usage = 
"Usage: pgmfft image.pgm\n"
"The results are stored in real.pgm and imag.pgm.\n"
;

#include <stdio.h>
#include <stdlib.h>
#include <complex.h>
#include <pgm.h>
#include <fftw3.h>

int main(int argc, char *argv[])
{
    gray maxval,outmaxval=255;
    gray** pixels;
    int xi,yi;
    int nx,ny;
    complex *in,*out;
    char* filename;

#define FOREACH(dothis) \
    for(yi=0;yi<ny;yi++) { for(xi=0;xi<nx;xi++) { dothis; } } 

    pgm_init(&argc,argv);     
    if(argc!=1+1) {
        fprintf(stderr,usage);
        exit(-1);
    }
    filename=argv[1];
    
    // Load the image
    {
        FILE* file = fopen(filename,"rb");
        if(!file) {
            fprintf(stderr,"Could not open %s for reading.\n");
            fprintf(stderr,usage);
            exit(-1);
        }
        pixels = pgm_readpgm(file,&nx,&ny,&maxval);
        fclose(file);
    }

    // Compute the FFT
    fftw_complex* ft = fftw_malloc(sizeof(fftw_complex)*nx*ny);         
    {
        fftw_complex* in = fftw_malloc(sizeof(fftw_complex)*ny*nx);         

        // Put the filter pixels into a flat array for fftw
        FOREACH(in[yi*nx+xi] = pixels[yi][xi]);
        
        fftw_plan plan = fftw_plan_dft_2d(ny, nx, in, ft, FFTW_FORWARD,
                FFTW_ESTIMATE);
        fftw_execute(plan);
        fftw_destroy_plan(plan);
        fftw_free(in);

        // Divide out factor of w*h.
        FOREACH(ft[yi*nx+xi] /= ny*nx);
    }

    // Compute min and max
    double rmin,rmax, imin,imax;
    {
        rmin=rmax=creal(ft[0]);
        imin=imax=cimag(ft[0]);
        FOREACH(
            complex val=ft[yi*nx+xi];
            rmin=MIN(rmin,creal(val)); rmax=MAX(rmax,creal(val));
            imin=MIN(imin,cimag(val)); imax=MAX(imax,cimag(val));
        );
    }

    // Output the real part, putting the min and max values in a comment at
    // the top.
    {
        char* outname = "real.pgm";
        FILE* outfile = fopen(outname,"wb");
        if(!outfile) {
            fprintf(stderr,"Could not open %s for writing.\n",outname);
            exit(-1);
        }
        fprintf(outfile,"P5\n");
        fprintf(outfile,"# min %f\n",rmin);
        fprintf(outfile,"# max %f\n",rmax);
        fprintf(outfile,"%i %i\n",nx,ny);
        fprintf(outfile,"%i\n",outmaxval);
        FOREACH(
            double val =
                outmaxval*(creal(ft[yi*nx+xi])-rmin)/(MAX(1.0,rmax-rmin));
            fputc((int)val,outfile)
            );
        fclose(outfile);
        fprintf(stderr,"Wrote %s\n",outname);
    }

    // Output the imaginary part, putting the min and max values in a comment at
    // the top.
    {
        char* outname = "imag.pgm";
        FILE* outfile = fopen(outname,"wb");
        if(!outfile) {
            fprintf(stderr,"Could not open %s for writing.\n",outname);
            exit(-1);
        }
        fprintf(outfile,"P5\n");
        fprintf(outfile,"# min %f\n",imin);
        fprintf(outfile,"# max %f\n",imax);
        fprintf(outfile,"%i %i\n",nx,ny);
        fprintf(outfile,"%i\n",outmaxval);
        FOREACH(
            double val =
                outmaxval*(cimag(ft[yi*nx+xi])-imin)/(MAX(1.0,imax-imin));
            fputc((int)val,outfile)
            );
        fclose(outfile);
        fprintf(stderr,"Wrote %s\n",outname);
    }

    return 0;
}

