// TODO:
// -v option
// Automatically save group cache files and reload them.
//
#include <complex.h>
#include <ppm.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <fftw3.h>
#include <assert.h>
#include <string.h>

#define MAX_IMAGES 1024

double square(double x) { return x*x; }

typedef struct _Image
{
    int ny,nx,nc;
    char filename[80];
    complex pxl[3][256*256]; // pixels
    complex dft[3][256*256]; // discrete fourier transform
    double  las[3][256*256]; // log of amplitude spectrum
}
Image;

Image* load_image(char filename[])
{
    int nx,ny, yi,xi,ci;
    Image* ret = calloc(1,sizeof(Image));
    double dmaxval;

    printf("Loading %s\n",filename);
    
    // Load the image
    pixel** pixels;
    {
        pixval maxval;
        FILE* file = fopen(filename,"rb");
        if(!file) {
            fprintf(stderr,"Could not open %s for reading\n",filename);
            return NULL;
        }
        strncpy(ret->filename,filename,sizeof(ret->filename));     
        pixels = ppm_readppm(file, &nx, &ny, &maxval);
        if(nx!=256) {
            fprintf(stderr,"Unsupported: Width of %s is not 256.\n",filename);
            return NULL;
        }
        if(ny!=256) {
            fprintf(stderr,"Unsupported: Height of %s is not 256.\n",filename);
            return NULL;
        }
        dmaxval=maxval;
        ret->nx=nx;
        ret->ny=ny;
        fclose(file);
    }

    // Transfer the pixels
    for(yi=0;yi<ny;yi++) {
        for(xi=0;xi<nx;xi++) {
            ret->pxl[0][yi*nx+xi] = pixels[yi][xi].r/dmaxval;
            ret->pxl[1][yi*nx+xi] = pixels[yi][xi].g/dmaxval;
            ret->pxl[2][yi*nx+xi] = pixels[yi][xi].b/dmaxval;
        }
    }
    
    // Compute Fourier transform, one per channel.
    for(ci=0;ci<3;ci++) {
        // This is wasteful and could be made more efficient by reusing the
        // plans.
        fftw_plan plan = fftw_plan_dft_2d(ny,nx, ret->pxl[ci], ret->dft[ci], 
                FFTW_FORWARD, FFTW_ESTIMATE);
        fftw_execute(plan);
        fftw_destroy_plan(plan);
    }

    // Compute the log of the amplitude spectrum on each channel.
    for(ci=0;ci<3;ci++) {
        for(yi=0;yi<nx;yi++) {
            for(xi=0;xi<nx;xi++) {
                complex z = ret->dft[ci][yi*nx+xi];
                ret->las[ci][yi*nx+xi] = z==0.0?-1e9:creal(clog(z));
            }
        }
    }

    ppm_freearray(pixels,ret->ny);
    return ret;
}


typedef struct 
{
    double mean[3][256*256]; // y, x, channel
    double sdev[3][256*256];
    Image* images[MAX_IMAGES];
    int num_images;
}
Group;

Group* load_group(char image_list_filename[])
{
    int i,ci,yi,xi, ny=256,nx=256;
    Group* g = calloc(1,sizeof(Group));
    FILE* file = fopen(image_list_filename,"r");
    if(!file) {
        fprintf(stderr,"Could not open %s for reading.\n", 
                image_list_filename);
        exit(-1);
    }
    for(i=0;i<MAX_IMAGES;i++) {
        char filename[4*80];
        if(!fgets(filename,sizeof(filename),file)) { break; }
        filename[strlen(filename)-1]='\0'; // strip trailing newline
        Image* img = load_image(filename);
        if(img) {
            g->images[g->num_images++] = img;
        }
    }
    fclose(file);

    if(g->num_images==0) {
        fprintf(stderr,
"No images were loaded from the filenames in %s.\n",image_list_filename);
        exit(-1);
    }
    
    // Compute the mean, mean of square, and sdev of las over the group
    for(ci=0;ci<3;ci++) { 
        for(yi=0;yi<ny;yi++) {
            for(xi=0;xi<nx;xi++) {
                double mean=0.0, msq=0.0,sdev;
                for(i=0;i< g->num_images;i++) {
                    Image* img = g->images[i];
                    double x = img->las[ci][yi*img->nx+xi];
                    mean += x;
                    msq += x*x;
                }
                mean /= g->num_images;
                msq /= g->num_images;
                sdev = msq - mean*mean;
                g->mean[ci][yi*256+xi] = mean;
                g->sdev[ci][yi*256+xi] = sdev;
            }
        }
    }

    return g;
}

// p(G1|img) where G1 is the proposition "img belongs to group 1."
double prob(Group* g1, Group* g2, Image* img)
{
    int i,ci,yi,xi, num_samples=256;
    double ratio,log_ratio=0.0;
    for(i=0;i<num_samples;i++) { 
        ci=drand48()*3;
        xi=drand48()*256;
        yi=drand48()*256;
        int index=yi*256+xi;
        // printf("g1 sdev: % 9.3g, g2 sdev: % 9.3g\n", g1->sdev[ci][index], g2->sdev[ci][index]);
        double las = img->las[ci][yi*256+xi];
        double sdevdiff = log(g2->sdev[ci][index])
                         -log(g1->sdev[ci][index]);
        double devdiff = 
            0.5*(square((las-g2->mean[ci][index])/g2->sdev[ci][index])
                -square((las-g1->mean[ci][index])/g1->sdev[ci][index]));
        double term = sdevdiff + devdiff;
        printf("sdevdiff: % 9.3g, devdiff: % 9.3g, term: % 9.3g\n",sdevdiff,devdiff,term);
        log_ratio += term;
    }
    printf("log ratio: %f\n",log_ratio);
    ratio=exp(log_ratio);
    return 1.0/(1.0+ratio);
}

int main(int argc, char *argv[])
{
    Group *g1,*g2;
    Image *img;
    double p;

    ppm_init(&argc,argv);

    if(argc!=1+1) {
            fprintf(stderr,
"\n"
"Usage: ppm-class-probability image.ppm\n"
"\n"
"There should be two files, one named group1 and the other named group2.\n"
"The file named group1 should contain the names of images in the first group,\n"
"and likewise for the file named group2.\n"
"The output of this program is a single number indicating the probability that\n"
"image.ppm belongs to the same class as the images listed in group1.\n"
"\n"
);
            exit(-1);
    }

    g1=load_group("group1");
    g2=load_group("group2");
    img=load_image(argv[1]);
    if(!img) { exit(-1); }
    p=prob(g1,g2,img);
    
    printf("%f\n",p);
    return 0;
}

