char* usage = 
"Usage: ppm-filter-response-rms [options] filter.pgm image.ppm\n"
"\n"
"Options\n"
"=======\n"
" -n : Pad filter with noise.  Default is to pad with zeros.\n"
" -v : Verify the result by computing it in the spatial domain.\n"
;

#include <complex.h>
#include <fftw3.h>
#include <math.h>
#include <ppm.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

double square(double x) { return x*x; }

int padding_with_noise=0;
int verification_requested=0;

int main(int argc, char *argv[])
{
    int yi,xi;
    char *filter_filename, *image_filename;
    pixval filt_maxval, img_maxval;
    int filt_nx,filt_ny, img_nx,img_ny, nx,ny;
    pixel** input_img_pixels;
    gray** input_filt_pixels;

    ppm_init(&argc,argv);

    // Parse arguments.
    {
        int bflag=0,ch;
        while ((ch=getopt(argc,argv,"hnv"))!=-1) {
            switch(ch) {
                case 'n': 
                    padding_with_noise=1;
                    break;
                case 'v': 
                    verification_requested=1;
                    break;
                case 'h':
                case '?':
                default:
                    fprintf(stderr,usage);
                    exit(-1);
            }
        }
        argc-=optind;
        argv+=optind;
        if(argc!=2) {
            fprintf(stderr,usage);
            exit(-1);
        }
        filter_filename=argv[0];
        image_filename=argv[1];
    }

    // Load the filter.
    {
        FILE* file=fopen(filter_filename,"rb");
        if(!file) {
            fprintf(stderr,"Could not open %s for reading.\n",
                    filter_filename);
            fprintf(stderr,usage);
            exit(-1);
        }
        input_filt_pixels=pgm_readpgm(file,&filt_nx,&filt_ny,&filt_maxval);
        fclose(file);
    }

    
    // Load the color image.
    {
        FILE* file=fopen(image_filename,"rb");
        if(!file) {
            fprintf(stderr,"Could not open %s for reading.\n",
                    image_filename);
            fprintf(stderr,usage);
            exit(-1);
        }
        input_img_pixels=ppm_readppm(file,&img_nx,&img_ny,&img_maxval);
        fclose(file);
    }

    nx=MAX(filt_nx,img_nx);
    ny=MAX(filt_ny,img_ny);

    // Compute the filter FFT
    fftw_complex* filt_dft = fftw_malloc(sizeof(fftw_complex)*ny*nx);         
    fftw_complex* filt = fftw_malloc(sizeof(fftw_complex)*nx*ny);         
    {
        int i,xi,yi;

        // Put the filter pixels into a flat array for fftw
        double s=0.0;
        for(yi=0;yi<ny;yi++) {
            for(xi=0;xi<nx;xi++) {
                double val = yi<filt_ny && xi<filt_nx
                                ?input_filt_pixels[yi][xi]
                                :padding_with_noise
                                    ?255*drand48()
                                    :0.0; // pad with zeros
                filt[yi*nx+xi] = val;
                s += val;
            }
        }

        fftw_plan plan = fftw_plan_dft_2d(ny, nx, filt, filt_dft,
                FFTW_FORWARD, FFTW_ESTIMATE);
        fftw_execute(plan);
        fftw_destroy_plan(plan);

        // Normalize the fft
        //
        // http://www.fftw.org/fftw3_doc/The-1d-Discrete-Fourier-Transform-_0028DFT_0029.html
        for(yi=0;yi<ny;yi++) {
            for(xi=0;xi<nx;xi++) {
                filt_dft[yi*nx+xi] /= sqrt(ny*nx);
            }
        }
    }

    // Compute the image FFT and RMS.
    fftw_complex* img_dft = fftw_malloc(sizeof(fftw_complex)*nx*ny);
    fftw_complex* img = fftw_malloc(sizeof(fftw_complex)*nx*ny);         
    {
        int xi,yi;

        // Put the image pixel luminances into a flat array for fftw
        for(yi=0;yi<ny;yi++) {
            for(xi=0;xi<nx;xi++) {
                double L = yi<img_ny && xi<img_nx 
                            ?PPM_LUMIN(input_img_pixels[yi][xi])/img_maxval
                            :padding_with_noise
                                ?drand48()
                                :0.0;
                img[yi*img_nx+xi] = L;
            }
        }

        fftw_plan plan = fftw_plan_dft_2d(ny, nx, img, img_dft,
                FFTW_FORWARD, FFTW_ESTIMATE);
        fftw_execute(plan);
        fftw_destroy_plan(plan);

        // Normalize the fft
        //
        // http://www.fftw.org/fftw3_doc/The-1d-Discrete-Fourier-Transform-_0028DFT_0029.html
        for(yi=0;yi<ny;yi++) {
            for(xi=0;xi<nx;xi++) {
                img_dft[yi*nx+xi] /= sqrt(ny*nx);
            }
        }
    }
    
    // Compute and output RMS of the pointwise product of their FFTs, which
    // is the same as the RMS of the spatial convolution, by the Fourier
    // convolution theorem and Plancherel's theorem.
    double rmsfrcifd=0.0; // RMS filter response computed in Fourier domain
    {
        int yi,xi;
        double normalized_rms;
        for(yi=0;yi<ny;yi++) {
            for(xi=0;xi<nx;xi++) {
                complex imgxy,fltxy;
                imgxy = img_dft[yi*nx+xi];  
                fltxy = filt_dft[yi*nx+xi]; 
                rmsfrcifd += square(cabs(imgxy*fltxy));
            }
        }
        rmsfrcifd=sqrt(rmsfrcifd/(nx*ny));
        printf("%f\n",rmsfrcifd);
    }

    if(verification_requested) {
        double rms=0.0;
        int yi,xi,fyi,fxi,yi2,xi2;
        for(yi=0;yi<ny;yi++) {
            for(xi=0;xi<nx;xi++) {
                double s=0.0; // s = convolution response at xi,yi
                for(yi2=0;yi2<ny;yi2++) {
                    for(xi2=0;xi2<nx;xi2++) {
                        s += filt[(((yi-yi2)+ny)%ny)*nx + ((xi-xi2)+nx)%nx]
                            *img[yi2*nx+xi2];
                    }
                }
                s/=sqrt(nx*ny);
                rms += s*s;
            }
        }
        rms = sqrt(rms/(nx*ny));
        printf("%f <- RMS response computed in spatial domain\n",rms);
        printf("Percent difference: %f\n",(rms-rmsfrcifd)/rmsfrcifd*100.0);
    }

    return 0;
}

