char *usage=
"summarize\n"
"a tool from the BrainMaps Analyze collection\n"
"written by Issac Trotts, 2006\n"
"\n"
"Usage1: summarize [options] command url-prefix\n"
"Usage2: summarize [options] -r resolution  url-prefix\n"
"\n"
"Options\n"
"=======\n"
"-a : Assemble the tiles into a complete image at each level.\n"
"-J : Do not remove downloaded jpeg tiles.\n"
"-r : Fetch a random tile and do nothing else.\n"
"\n"
"For example, the following command generates a small summary image showing\n"
"the entropy of tiles, do this:\n"
"\n"
" summarize ppmentropy \\\n"
"   http://brainmaps.org/HBP2/m.mulatta/RH04/RH04-highres/0501\n"
"\n"
"This example downloads a random tile at resolution 9:\n"
"\n"
" summarize -r 9 http://brainmaps.org/HBP2/m.mulatta/RH04/RH04-highres/0501\n"
"\n";

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>
#include <unistd.h>
#include <ctype.h>
#include <ppm.h>

#ifndef min
#define min(a,b) ((a)<(b)?(a):(b))
#endif

#ifndef max
#define max(a,b) ((a)>(b)?(a):(b))
#endif

typedef struct 
{
    int group_index;
}
Tile;

typedef struct
{
    int num_rows;
    int num_cols;
    Tile **tiles; /* [row][colummn] */
}
ResLevel;

static ResLevel *levels=NULL;
static int full_width=0; 
static int full_height=0;
static int num_tiles=0;
static int tile_size=256; 
static int num_resolutions=0;
static char *url_prefix=NULL;
static int verbosity=2;
static int removing_jpeg_tiles=1;
static int will_fetch_random_tile=0;
static FILE* tilestatsfile=NULL;

void my_system(char *cmd)
{
    if(system(cmd)!=0) {
        fprintf(stderr,"Command failed: %s\n",cmd);
        exit(-1);
    }
}


void expect_string(char str[], FILE* file, char filename[])
{
    int c,i=0, bufsize=80;
    char *s, *buf;
    buf = malloc(bufsize);
    for(s=str; *s!='\0'; s++) {
        if(i+2>bufsize) {
            bufsize*=2;
            buf=realloc(buf,bufsize);
        }
        c=fgetc(file);
        buf[i++]=c;
        buf[i]='\0';
        if(c!=*s) {
            fprintf(stderr, "Error: Expected %s but got %s in %s\n",
                    str,buf,filename);
            exit(-1);
        }
    }
    free(buf);
}

void load_image_properties_file(void)
{
    int num_images, expected_num_tiles;
    char cmd[3*80];
    char *filename = "ImageProperties.xml";
    FILE *file;
    snprintf(cmd,sizeof(cmd),"rm -f %s",filename);
    my_system(cmd);
    snprintf(cmd,sizeof(cmd),"wget %s/%s",url_prefix,filename);
    fprintf(stderr,"%s\n",cmd);
    if(system(cmd)!=0) {
        fprintf(stderr,"Could not fetch %s from the server.\n",filename);
        exit(-1);
    }
    file=fopen(filename,"r");
    if(file==NULL) {
        fprintf(stderr,"Could not open %s for reading.\n",filename);
        exit(-1);
    }

    expect_string("<IMAGE_PROPERTIES WIDTH=\"",file,filename);
    fscanf(file,"%i",&full_width);
    expect_string("\" HEIGHT=\"",file,filename);
    fscanf(file,"%i",&full_height);
    expect_string("\" NUMTILES=\"",file,filename);
    fscanf(file,"%i",&num_tiles);
    expect_string("\" NUMIMAGES=\"",file,filename);
    fscanf(file,"%i",&num_images);
    expect_string("\" VERSION=\"1.8\" TILESIZE=\"",file,filename);
    fscanf(file,"%i",&tile_size);
    fclose(file);

    if(tile_size!=256) {
        fprintf(stderr,"Unsupported tile size: %i != 256.\n",tile_size);
        exit(-1);
    }

    expected_num_tiles = (full_width/tile_size + 1)*(full_height/tile_size + 1);
    if(expected_num_tiles != num_tiles) {
        fprintf(stderr,
                "Warning: expected_num_tiles is %i, but num_tiles is %i\n",
                expected_num_tiles, num_tiles);
    }

    num_resolutions = 
        1+ceil(log(ceil(max(full_width,full_height)/tile_size))/log(2.0));
    if(verbosity>=1) { 
        fprintf(stderr,"There are %i levels of resolution.\n",num_resolutions);
    }
}

void fetch_tile(int res, int row, int col)
{
    // Fetch the tile from the server.
    char url[3*80];
    char cmd[4*80];
    char jpegname[80];
    struct timeval t0,t1;
    struct stat my_stat;
    float elapsed_msec;
    static int have_warned=0;
    Tile* tile = &levels[res].tiles[row][col];
    char basename[80];
    snprintf(basename,sizeof(basename),"%i-%i-%i",res,col,row);
    snprintf(url,sizeof(url),"%s/tilegroup%i/%s.jpg",
            url_prefix,tile->group_index,basename);
    snprintf(cmd,sizeof(cmd),"wget -c -nv %s",url);
    snprintf(jpegname,sizeof(jpegname),"%s.jpg",basename);
    gettimeofday(&t0,NULL);
    my_system(cmd);
    gettimeofday(&t1,NULL);
    elapsed_msec = (t1.tv_sec - t0.tv_sec)*1000.0
        + (t1.tv_usec - t0.tv_usec)/1000.0;
    if(will_fetch_random_tile) {
        printf("%s\n",jpegname);
    }
    if(tilestatsfile) {
        if(!stat(jpegname, &my_stat) && !have_warned) {
            fprintf(stderr,
    "Warning: stat failed on %s.  Future warnings suppressed.\n",jpegname);
            have_warned=1;
        }
        fprintf(tilestatsfile,"%100s\t%10f\t%10f\n",
                url,elapsed_msec,my_stat.st_size/1000.0);
        fflush(tilestatsfile);
    }
}

int main(int argc, char *argv[])
{
    int assembling_images=0, requested_res=1;
    char* command;

    ppm_init(&argc,argv);

    // Parse command line arguments
    {
        int bflag=0, ch, fd;
        while ((ch=getopt(argc,argv,"aJr"))!=-1) {
            switch(ch) {
                case 'a': assembling_images=1; break;
                case 'J': removing_jpeg_tiles=0; break;
                case 'r': will_fetch_random_tile=1; break;
                case '?':
                default:
                    fprintf(stderr,usage);
                    exit(-1);
            }
        }
        argc-=optind;
        argv+=optind;
        if(will_fetch_random_tile) {
            if(argc!=2) {
                fprintf(stderr,"Please specify a URL prefix.\n");
                fprintf(stderr,usage);
                exit(-1);
            }
            requested_res=strtol(argv[0],NULL,10);
            url_prefix=argv[1];
        }
        else if(argc==2) {
            command=argv[0];
            url_prefix=argv[1];
        }
        else {
            fprintf(stderr,usage);
            exit(-1);
        }
    }

    load_image_properties_file();

    /* Allocate the tile array, for all levels. */
    {
        int res,row;
        levels = calloc(num_resolutions,sizeof(levels[0]));
        for(res=0;res<num_resolutions;res++) {
            levels[res].tiles = calloc(1<<res,sizeof(levels[res].tiles[0]));
            levels[res].num_rows=
                ceil(full_height/(256.0*pow(2.0,num_resolutions-res-1)));
            levels[res].num_cols=
                ceil(full_width/(256.0*pow(2.0,num_resolutions-res-1)));
            for(row=0; row<levels[res].num_rows; row++) {
                levels[res].tiles[row] = 
                    calloc(levels[res].num_cols,
                            sizeof(levels[res].tiles[row][0]));
            }
        }
    }

    /* Figure out the tile group and id of each tile. */
    {
        int res,row,col,tile_count=0;
        for(res=0; res<num_resolutions; res++) {
            for(row=0; row<levels[res].num_rows; row++) {
                for(col=0; col<levels[res].num_cols; col++) {
                    Tile* tile = &levels[res].tiles[row][col];
                    tile->group_index = tile_count/256;
                    tile_count++;
                }
            }
        }
    }

    if(will_fetch_random_tile) {
        int row,col;
        struct timeval t;
        gettimeofday(&t,NULL);
        srand48(t.tv_usec);
        if(requested_res>=num_resolutions) {
            fprintf(stderr,
"Warning: requested resolution %i is too high.  Setting it to %i.\n",
requested_res,num_resolutions-1);
            requested_res=num_resolutions-1;
        }
        row=drand48()*levels[requested_res].num_rows;
        col=drand48()*levels[requested_res].num_cols;
        fetch_tile(requested_res,row,col);
        exit(0);
    }

    /* Create a directory where the results will be stored, and cd into it. */
    {
        char* c;
        int status;
        char outdirname[4*80];

        // Make the dirname a combination of command and url prefix.
        snprintf(outdirname,sizeof(outdirname),"%s--%s",
                url_prefix,
                // remove ./ if present
                command[0]=='.'&&command[1]=='/'?command+2:command); 

        // Convert dots, colons, slashes, and spaces to dashes.
        for(c=outdirname; *c; c++) { 
            *c = isspace(*c)||*c=='-'||*c=='.'||*c=='/'||*c==':'
                ?'-'
                :*c;
        }

        // Make the directory.
        if((status=mkdir(outdirname,S_IRWXU|S_IRWXG|S_IROTH|S_IXOTH))!=0) {
            if(errno!=EEXIST) {
                fprintf(stderr,"Failed to create output directory %s",
                        outdirname);
                perror("");
                exit(-1);
            }
        }

        if(chdir(outdirname)!=0) {
            fprintf(stderr,"Could not chdir into %s\n",outdirname);
            exit(-1);
        }
    }

    // Open the tile times file. 
    char* tilestatsfilename = "tilestats.txt";
    tilestatsfile = fopen(tilestatsfilename,"w");
    if(!tilestatsfile) {
        fprintf(stderr,"Could not open %s for writing.\n",tilestatsfilename);
        exit(-1);
    }
    else {
        fprintf(stderr,"Tile fetch times will be stored in %s.\n",tilestatsfilename);
    }

    /* Construct new images, one per res, by marching through the tiles,
     * running the analysis command on each one. */
    {
        int res,row,col;
        for(res=1;res<num_resolutions;res++) {
            int w,h;
            FILE* outfile;
            float *pixels;
            pixel** a_pixels; // pixels in assembled image, if any
            w=levels[res].num_cols;
            h=levels[res].num_rows;
            pixels=malloc(w*h*sizeof(pixels[0]));
            a_pixels = assembling_images?ppm_allocarray(w*256,h*256):NULL;
            for(row=0;row<h;row++) {
                for(col=0;col<w;col++) {
                    char basename[80];
                    char cmd[2*80];
                    Tile* tile;

                    tile = &levels[res].tiles[row][col];
                    snprintf(basename,sizeof(basename),"%i-%i-%i",res,col,row);

                    fetch_tile(res,row,col);

                    // Convert the downloaded tile from JPEG to PPM.
                    {
                        char cmd[2*80];
                        snprintf(cmd,sizeof(cmd),"djpeg %s.jpg > %s.ppm",
                                basename,basename);
                        my_system(cmd);
                    }

                    // Run the analysis command on the tile
                    char* result_name="result";
                    {
                        snprintf(cmd,sizeof(cmd),"%s %s.ppm > %s",
                                command,basename,result_name);
                        my_system(cmd);
                    }

                    // Load the result of the analysis command. 
                    float result_val;
                    {
                        FILE* infile=fopen(result_name,"r");
                        if(!infile) {
                            fprintf(stderr,"Could not open %s for reading.\n",
                                    result_name);
                            exit(-1);
                        }
                        if(fscanf(infile,"%f",&result_val)!=1) {
                            fprintf(stderr,
"Could not read the expected floating point number in %s.  Using noise.\n",
result_name);
                            result_val = drand48();
                        }
                        fclose(infile);
                    }

                    // Save the result in a pixel of the output image.
                    pixels[row*w+col] = result_val;

                    // Possibly add the tile to the assembled image.
                    if(assembling_images) {
                        FILE* file; 
                        char ppm_tile_filename[80];
                        static have_warned=0;
                        snprintf(ppm_tile_filename,sizeof(ppm_tile_filename),
                                "%s.ppm",basename);
                        file = fopen(ppm_tile_filename,"r");
                        if(!file && !have_warned) {
                            fprintf(stderr,
"Could not load %s.  Future warnings will be suppressed.\n");
                            have_warned=1;
                        }
                        else {
                            pixel** tile_pixels;
                            pixval tile_maxval;
                            int yi,xi, tile_cols,tile_rows;
                            tile_pixels = ppm_readppm(file,&tile_cols,
                                    &tile_rows,&tile_maxval);
                            for(yi=0;yi<tile_rows;yi++) {
                                for(xi=0;xi<tile_cols;xi++) {
                                    a_pixels[tile_size*row+yi][tile_size*col+xi] = 
                                        tile_pixels[yi][xi];
                                }
                            }
                            ppm_freearray(tile_pixels,tile_rows);
                            fclose(file);
                        }
                    }

                    // Remove the PPM version of the tile.
                    {
                        char cmd[2*80];
                        snprintf(cmd,sizeof(cmd),"rm %s.ppm",basename);
                        if(system(cmd)!=0) {
                            fprintf(stderr,
"Warning: Could not remove %s.ppm\n",basename);
                        }
                    }

                    // Remove the JPEG version of the tile.
                    if(removing_jpeg_tiles)
                    {
                        char cmd[2*80];
                        snprintf(cmd,sizeof(cmd),"rm %s.jpg",basename);
                        if(system(cmd)!=0) {
                            fprintf(stderr,
"Warning: Could not remove %s.jpg\n",basename);
                        }
                    }
                }
            }

            if(assembling_images) {
                // Save the assembled image for this level.
                {
                    FILE* file;
                    char a_outname_ppm[80];
                    snprintf(a_outname_ppm,sizeof(a_outname_ppm),
                            "res%02i-assembled.ppm",res);
                    file=fopen(a_outname_ppm,"wb");
                    if(!file) {
                        fprintf(stderr,"Could not open %s for writing.\n",
                                a_outname_ppm);
                    }
                    else {
                        ppm_writeppm(file,a_pixels,w*256,h*256,255,0);
                        fclose(file);
                    }
                }

                // Convert it to a JPEG.
                {
                    char cmd[4*80];
                    snprintf(cmd,sizeof(cmd),
"cjpeg -q 90 res%02i-assembled.ppm > res%02i-assembled.jpg",res,res);
                    // If cjpeg succeeds, let's get rid of the probably very
                    // large ppm image.
                    if(system(cmd)==0) {
                        snprintf(cmd,sizeof(cmd),"rm res%02i-assembled.ppm",res);
                        system(cmd);
                    }
                }
            }

            // Rescale the output image values.
            {
                // Compute min and max.
                float mn=pixels[0],mx=pixels[0];
                {
                    int i;
                    for(i=0;i<w*h;i++) { 
                        mn=min(mn,pixels[i]);
                        mx=max(mx,pixels[i]);
                    }
                    mx=max(mx,mn+1e-3); // Deal with a possibly constant image.
                }
                
                // Rescale to [0,1].
                {
                    int i;
                    for(i=0;i<w*h;i++) { 
                        pixels[i]=(pixels[i]-mn)/(mx-mn);
                    }
                }
            }
            
            // Save the output image
            char outbasename[80];
            {
                int row,col;
                char outname[80+4];
                snprintf(outbasename,sizeof(outbasename),"res%i",res);
                snprintf(outname,sizeof(outname),"%s.ppm",outbasename);
                outfile=fopen(outname,"wb");
                if(!outfile) {
                    fprintf(stderr,"Could not open %s.ppm for writing.\n",
                            outbasename);
                    exit(-1);
                }
                fprintf(outfile,"P6\n");
                fprintf(outfile,"%i %i\n",w,h);
                fprintf(outfile,"255\n");
                for(row=0;row<h;row++) {
                    for(col=0;col<w;col++) {
                        int r,g,b;
                        r=g=b=pixels[row*w+col]*255;
                        fputc(r,outfile);
                        fputc(g,outfile);
                        fputc(b,outfile);
                    }
                }
                fclose(outfile);
                fprintf(stderr,"Wrote %s.ppm\n",outbasename);
            }

            // Convert the output image to a JPEG for easy viewing.
            {
                char cmd[2*80];
                static int have_warned=0;
                snprintf(cmd,sizeof(cmd),"cjpeg %s.ppm > %s.jpg",
                        outbasename,outbasename);
                if(system(cmd)!=0 && !have_warned) {
                    fprintf(stderr,"Command failed: %s\n",cmd);
                    fprintf(stderr,"Future warnings suppressed.\n",cmd);
                    have_warned=1;
                }
                else {
                    fprintf(stderr,"Wrote %s.jpg\n",outbasename);
                }
            }

            free(pixels);
        }
    }

    fclose(tilestatsfile);

    system("rm -f result ImageProperties.xml");

    return 0;
}


